import java.util.*;

/**
 * Implements a deterministic Turing Machine (TM) with a single, left-bounded tape
 */
public class TuringMachine {
	// Integer constants for head movements
	public static final int HOLD = 0, LEFT = 1, RIGHT = 2;
	
	// Tabular representation of deterministic transition matrix
	private TMResult[][] transitionMatrix;
	private TMTape tape;
	private TMState currentState;
	private int currentPosition;
	private Set states, symbols;
	// Utility data structure to map names to instances for states and symbols
	private Map nameToState, nameToSymbol;
	private int numStates, numSymbols;
	
	// Create a new instance for a TM with maximal number of states and symbols
	public TuringMachine(int maxStates, int maxSymbols) {
		tape = new TMTape();
		states = new HashSet(); symbols = new HashSet();
		nameToState = new HashMap(); nameToSymbol = new HashMap();
		numStates = 0; numSymbols = 0;
		transitionMatrix = new TMResult[maxStates][maxSymbols];
	}
	
	// Retrieve tape for preparing of input
	public TMTape getTape() {
		return tape;
	}

	// Add a new state to the TM if it does not already exist
	public TMState addState(String name, boolean isStart) {
		TMState s = (TMState) nameToState.get(name);
		if (s == null) {
			s = new TMState(name, isStart, numStates);
			states.add(s);
			nameToState.put(name, s);
			numStates++;
		}
		return s;
	}
	
	// Add a new symbol to the TM if it does not already exist
	public TMSymbol addSymbol(String symbol) {
		TMSymbol s = (TMSymbol) nameToSymbol.get(symbol);
		if (s == null) {
			s = new TMSymbol(symbol, numSymbols);
			symbols.add(s);
			nameToSymbol.put(symbol, s);
			numSymbols++;
		}
		return s;
	}
		
	// Add a new transition to the TM (all referenced states & symbols have to exist already)
	public void addTransition(String fromState, String inputSymbol, 
							  String toState, String outputSymbol, int headDirection) {
		TMState origin = (TMState) nameToState.get(fromState);
		TMState target = (TMState) nameToState.get(toState);
		TMSymbol input = (TMSymbol) nameToSymbol.get(inputSymbol);
		TMSymbol output = (TMSymbol) nameToSymbol.get(outputSymbol);
		TMResult result = new TMResult(target, output, headDirection);
		transitionMatrix[origin.getId()][input.getId()] = result;
	}
	
	// Initialize TM with start state and given position on the tape
	public void init(int pos) {
		currentPosition = pos;
		Iterator stateIt = states.iterator();
		while (stateIt.hasNext()) {
			TMState s = (TMState) stateIt.next();
			if (s.isStart()) {
				currentState = s;
				break;
			}
		}
	}
	
	// Execute TM after previous initialisation, providing trace output of current state,
	// position and tape content if desired
	public void execute(boolean doTrace) {
		boolean running = true;
		while (running) {
			if (doTrace) {
				for (int i = 0; i < currentPosition; i++)
					System.out.print("  ");
				System.out.print(" " + currentState.getId() + "\n");
				System.out.println(tape);
			}
			running = step();
		}
	}
	
	// Perform a singe step of the TM, returns true if the TM may continue or false otherwise
	private boolean step() {
		TMSymbol symbol = tape.read(currentPosition);
		TMResult result = transitionMatrix[currentState.getId()][symbol.getId()];
		if (result != null) {
			currentState = result.getState();
			tape.write(currentPosition, result.getSymbol());
			switch (result.getHeadDirection()) {
				case LEFT: currentPosition--; break;
				case RIGHT: currentPosition++; break;
				case HOLD: ; // Do nothing
			}
			return true;
		}
		return false;
	}
}
